/*
	  This file is part of proteinortho.
	  (C) 2009 Marcus Lechner
 
	  proteinortho is free software; you can redistribute it and/or modify
	  it under the terms of the GNU General Public License as published
	  by the Free Software Foundation; either version 2, or (at your
	  option) any later version.
 
	  proteinortho is distributed in the hope that it will be useful, but
	  WITHOUT ANY WARRANTY; without even the implied warranty of
	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	  General Public License for more details.
 
	  You should have received a copy of the GNU General Public License
	  along with proteinortho; see the file COPYING.  If not, write to the
	  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
	  Boston, MA 02111-1307, USA.	
*/

/*
 *  ConnectedComponentFinder - core part of proteinortho
 *  finds connected components out of a list of edges
 *
 *  if two or more nodes within a connected component belong to the same
 *  species, they are marked as paralog
 *
 *  @author Marcus Lechner, Lydia Steiner
 *  @email marcus@bioinf.uni-leipzig.de
 *  @company Bioinformatics, University of Leipzig
 *  @version 1.00
 *  @date 19-02-2009
 */


#include <stdio.h>
#include <string.h>
#include "stack.h"
#include "basic-types.h"

Uint** read_graph(Uint** graph);
void ccfinder(Uint** graph, Uint* gen_vieh, Uint size, Uint viecher);
Uint* read_gvmapping(Uint* liste, Uint* size, Uint* max);

int main(void) {
	/* allocate memory */
	Uint size = 10000000;
	Uint* gen_vieh = (Uint*) malloc(sizeof(Uint)*size);
	Uint viehnum = 0;

	/* read the gene -> species mapping */
	/* pointer to gen_vieh might be moved, so set him to the new location */
	fprintf(stderr," >Loading species vs. gene mapping...");
	gen_vieh = read_gvmapping(gen_vieh,&size,&viehnum);
	/* gene_vieh: gen_vieh[gene-no.] -> species-no. */
	fprintf(stderr,"done\n");

	/* allocate memory for the graph */
	Uint** graph = (Uint**) malloc(sizeof(Uint*)*size);

	/* init all fields with 0 */
	memset(graph, 0, sizeof(Uint*)*size);

	/* load the graph via an edge-list */
	fprintf(stderr," >Loading graph...");
	graph = read_graph(graph);
	/* graph[start-gene-no.] contains:
	 *  0				1		2		...	[n = no. of entries]
       * [no. of entries]	[target_1]	[target_2]	...	[target_n]
	 *
	 * every entry is reflexiv, so if we have an edge 2->3 there is also one 3->2	
       */
	fprintf(stderr,"done\n");

	/* run the main-algorithm */
	fprintf(stderr," >Running connected component analysis...");
	ccfinder(graph, gen_vieh, size, viehnum);
	fprintf(stderr,"done\n");

	/* free */
	free(gen_vieh);
	int i;
	for(i = 0; i < size; i++){free(graph[i]);}
	free(graph);
      return 0;
}

/*  
 *  main algorithm, finds connected components and prints them to output
 *  
 *  graph	- graph represented as an edgelist
 *  gen_vieh- tells which genes belong to which species
 *  size	- number of genes (= proteines)
 *  viecher	- number of species
 *  return	- nothing, but prints data to stdout   
*/
void ccfinder(Uint** graph, Uint* gen_vieh, Uint size, Uint viecher) {
	/* init the stack */
	Stack stack;
	bl_stackInit(&stack, viecher);
	
	/* color of the first connected component */
	int ccnum = 1;

	/* list of species that are within the current connected component */
	Uint hamma[viecher];

	/* list of proteins within the current connected component */
	Uint schoninnercc[size];

	/* allocate memory for output list */
	int** gibmichaus = (int**) malloc(sizeof(int*)*viecher);

	/* some loop-variables */
	int diana;
	int marcus;
	int lydia;
	int erik;	
	Uint  j;

	/* zero-init species list, output list */
	for(marcus = 0; marcus < viecher; marcus++){
		hamma[marcus] = 0;
		gibmichaus[marcus] = (int*) malloc(0);
	}
	/* and protein list */
	for(marcus = 0; marcus < size; marcus++){
		schoninnercc[marcus] = 0;
	}	

	/* foreach protein */
	for (erik = 0; erik < size; erik++) {
		/* we have not found a paralog protein, for we just started */
		int paralog_found = 0;

		/* skip protein, if already colored */
		if(schoninnercc[erik] != 0) continue;

		/* else add it to the stack */
		bl_stackPush(&stack, erik);				

		/* and colorize it, so it won't be used again */
		schoninnercc[erik] = ccnum;				

		/* while there are proteins on the stack */
		while(!bl_stackIsEmpty(&stack)){

			/* fetch the first from */
			Uint akt = bl_stackPop(&stack);
			
			/* note, that we've one protein of this species more */
			hamma[gen_vieh[akt]]++;

			/* allocate enlarged space for output */
			gibmichaus[gen_vieh[akt]] = (int*) realloc(gibmichaus[gen_vieh[akt]],sizeof(int)*hamma[gen_vieh[akt]]);

			/* and store the current protein there */
			gibmichaus[gen_vieh[akt]][hamma[gen_vieh[akt]]-1] = (int) akt;

			/* if the species was already included, note that */
			if(hamma[gen_vieh[akt]] > 1) paralog_found = 1;		
				
			/* push all neighbors on the stack */
			/* foreach edge */
			for (j = 1; j < graph[akt][0]; j++) {			
				/* if not yet colored */
				if(schoninnercc[graph[akt][j]] == 0) {
					/* add to stack */		
					bl_stackPush(&stack, graph[akt][j]);
					/* and colorize */	
					schoninnercc[graph[akt][j]] = ccnum;	
				}
			}
		}
		/* choose another color for the next component, this has no effect at the moment
               but might be usefull later */
		// ccnum++;

		/* print connected component to output */
		/* foreach species */
		for(lydia = 0; lydia < viecher; lydia++){
			/* if no gene found for it, print * */
			if(hamma[lydia] == 0)printf("*");

			/* else */
			else {
				/* print it's number */
				printf("%d",gibmichaus[lydia][0]);

				/* of there are more proteins for this species */
				for (diana = 1; diana < hamma[lydia]; diana++) {
					/* print them also but comma-seperated */
					printf(",%d",gibmichaus[lydia][diana]);															
				}
			}
			/* if we reached the last species */
			if(lydia == viecher-1) {
				/* add the paralog note, if there are paralogs in */
				if (paralog_found == 1) printf("\tparalog");
				/* finish line */
				printf("\n");
			}
			/* else print a tab */
			else printf("\t");
		}

		/* zero-init output list */
		for(diana = 0; diana < viecher; diana++){
			gibmichaus[diana] = (int*) realloc(gibmichaus[diana],0);
			hamma[diana] = 0;
		}
	}

	/* free species list and output list for the next protein */
	for(diana = 0; diana < viecher; diana++){
		free(gibmichaus[diana]);
	}
	free(gibmichaus);
	
	/* free stack */	
	bl_stackDestruct(&stack);
}


/* 
 *  load the genes-species mapping from file
 *  
 *  liste	- initially allocated array for genes-species mapping
 *  size	- number of genes (= proteines), here: initial size of list
 *		  (will be set here to the actual no. of genes)
 *  max	- number of species found (will be set here)
 *  returns	- new pointer for liste (it may haven been realloced to somewhere different)
 *
 *  file format:
 *  gene-no. species-no.
 */
Uint* read_gvmapping(Uint* liste, Uint* size, Uint* max) {
	/* open the mappingfile */
	FILE* datei = fopen ("cout.gvmapping", "r");
	Uint a, b, counter = 0;
	
	/* read through the file */
	while(fscanf (datei, "%d %d\n", &a, &b) != EOF) {
		/* count lines */
		counter++;
		
		/* if gene-no. is bigger than the vector */
		if (a >= *size) {
			/* increase it's size */
			*size = 10000000+a;
			/* and realloc accordingly */
			liste = (Uint*) realloc(liste,sizeof(Uint)*(*size));
		}
		
		/* store the data */
		/* liste[gen] = species */
		liste[a] = b;

		/* hold the maximum species-no. */
		if(*max < b) *max = b;
	}

	/* close file */
	fclose (datei);

	/* reallocate list-size to what is actualy needed */
      /* previously we increased it's size in big steps, so it might be to big */
	liste = (Uint*) realloc(liste,sizeof(Uint)*counter);

	/* for max is ment to be the border for array-indexing, increase it by one */
	*max += 1;
	/* set size to the actual no. of genes found */
	*size = counter;
	
	/* return the fetched list */
	return liste;
}

/*  
 *  load the graph (corresponding to blasthits) from file
 *  
 *  graph	- graph represented as an edgelist (will be set here)
 *  returns	- new pointer for graph (it may haven been realloced to somewhere different)
 *  
 *  file format:
 *  gene-no.1 gene-no.2
 */
Uint** read_graph(Uint** graph) {

	/* open the edge-list file */
	FILE* datei = fopen ("cout", "r");

	/* dummy to reconstruct the numbers, none of them should have more than 100 signs */
	char word[100];

	/* start reading the file initaly */
	char c = fgetc(datei);

	/* init the other variables */
	int counter = 0, schalter = 0;
	Uint value = 0, index = 0, sizecounter = 0;

	/* read through the file, char by char */
	while (c != EOF) {
		/* if the current character is a space or new line */
		if (c == ' ' || c == '\n')  {
			/* finish the word there, to make atoi work */
			word[counter] = '\0';
			
			/* gather the value from number as char-array */
			value = (Uint) atoi(word);

			/* set counter 0 for the next word */
			counter = 0;

			/* if it is the first number of a line */
			if (schalter == 0) {
				/* the edge starts at this gene (= captured number) */
				index = value;
				
				/* indicate switch to frist step has been done */
				schalter++;

				/* if this targetlist has allready been initialaised */
                        /* (the might be more than one target if protheinorto is used fuzzy) */				
				if(graph[index] != NULL) {
					/* set size to what it was before +1 */
					sizecounter = graph[index][0] + 1;

					/* set switch to the new value */
					schalter = graph[index][0];
	
					/* allocate memory for the new edge */
					graph[index] = (Uint*) realloc(graph[index],sizeof(Uint)*sizecounter);
				}
				/* if not, it's the first entry */
				else {
					/* initialise it */					

					/* allocate needed memory */
					graph[index] = (Uint*) malloc(sizeof(Uint));

					/* we have one entry here */
					sizecounter = 1;

					/* set no. of entries to 1 (for at least one will be stored) */
					graph[index][0] = 1;
				}
			}
			/* we handle an edge target, not a start */
			else {
				/* the switch indicates the count of numbers in the current line
                           if these are more than we have space allocated... */
				if (schalter >= sizecounter) {
					/* ... we need to increase size */
					sizecounter++;
					
					/* and reallocate the target-list accordingly */
					graph[index] = (Uint*) realloc(graph[index],sizeof(Uint)*sizecounter);
				}
				
				/* store the target and increase the switch */
				graph[index][schalter++] = value;
			
				/* store the number of targets in list */
				graph[index][0] = schalter;

				/* all edges in the file are bidirectional, to save 1/2 filesize
                           so we have to do it vice versa with start and target swapped */

				/* if this targetlist has allready been initialaised */
				if (graph[value] != NULL) {
					/* reallocate one more space for the target list */
					graph[value] = (Uint*) realloc(graph[value],sizeof(Uint)*(graph[value][0]+1));

					/* store the new target in the list */
					graph[value][graph[value][0]] = index;

					/* we have one more entry now, note that */
					graph[value][0]++;
				}
				/* if not, it's the first entry */
				else {
					/* allocate memory for two numbers, for we know the will be a first now */
					graph[value] = (Uint*) malloc(sizeof(Uint)*2);

					/* store the information that we have 2 entries */					
					graph[value][0] = 2;

					/* store the target in the list */					
					graph[value][1] = index;
				}
			}

			/* if the char is a new line, the egde list entry has finished here */
			if (c == '\n') {
				/* reinit the switch (= number of numbers in the line) to 0 */
				schalter = 0;
			}
		}
	
		/* in any othe case the next char is a number */
		else {
			/* add it's sign to word */
			word[counter++] = c;
		}

		/* read the next char */
		c = fgetc(datei);
	}
 
	/* close file */
	fclose (datei);

	/* return the graph-pointer */
	return graph;
}
