#!/usr/bin/perl

##########################################################################################
#	  This file is part of proteinortho.
#	  (C) 2009 Marcus Lechner
# 
#	  proteinortho is free software; you can redistribute it and/or modify
#	  it under the terms of the GNU General Public License as published
#	  by the Free Software Foundation; either version 2, or (at your
#	  option) any later version.
#
#	  proteinortho is distributed in the hope that it will be useful, but
#	  WITHOUT ANY WARRANTY; without even the implied warranty of
#	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	  General Public License for more details.
#
#	  You should have received a copy of the GNU General Public License
#	  along with proteinortho; see the file COPYING.  If not, write to the
#	  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#	  Boston, MA 02111-1307, USA.	
##########################################################################################

##########################################################################################
# About
##########################################################################################
# orthomatrix2tree
# input ortheset output from proteinortho
# output corresponding UPGMA tree in newick format
#
# branchlabels show the number of common genes in the subtrees
# branchlengths represent the number of new common genes since the last node
# 
# @author Marcus Lechner
# @email marcus@bioinf.uni-leipzig.de
# @company Bioinformatics, University of Leipzig
# @version 3.00
# @date 19-02-2009
#
##########################################################################################

##########################################################################################
# Imports
##########################################################################################
use strict;
use warnings "all";
use File::Basename;


##########################################################################################
# Usage, defaults and parameters
##########################################################################################
my $usage= << "JUS";
  usage:  orthomatrix2tree.pl ORTHOMATRIX >OUTTREE
  input:  output from proteinortho
  output: corresponding UPGMA tree in newick format
JUS

# path to the C-part of this program
my @tmppath = fileparse($0);
our $scriptpath = $tmppath[1];

# open the given file, this should be a matrix
open(SOURCE,"<$ARGV[0]") || die("Could not open file $ARGV[0]!\n$usage");
open(TRG,">$ARGV[0].tmp.matrix") || die("Could not open tmp-file!");
my $did = 0;
my @viecher = ();

# convert the orthoset to a 1/0 matrix
print STDERR "Preparing datafile...\n";
while(<SOURCE>) {
	my @data = split;
	if ($data[0] =~ /^\#/) {
		# first line
		if ($did == 0) {
			shift(@data);
			# species names to header
			print TRG scalar(@data),"\n";
			@viecher = @data;
			$did = 1;
		}
		# second line
		elsif ($did == 1) {
			shift(@data);
			for (my $i = 0; $i < scalar(@data); $i++) {
				$viecher[$i] =~ s/.*\///g;
				$viecher[$i] =~ s/\..*$//g;
				# if length is bigger than 230 signs, cut
				# (Treebuilder in C can only operate 255 chars including genecount)
				if (length($viecher[$i]) > 230) {
					$viecher[$i] = substr($viecher[$i],0,230);
				}
				# add gene count to names and print to header
				print TRG $viecher[$i],"_[",$data[$i],"] ";
			}
			print TRG "\n";
			$did = 2;
		}
		next;
	}

	# any futher line
	foreach (@data) {
		# just transform the entries to 0 and 1
		# the gene-names are not important here
		if ($_ eq "*") {
			print TRG 0;
		}
		else {
			print TRG 1;
		}
		print TRG " ";
	}
	print TRG "\n";
}
close(SOURCE);
close(TRG);

# Run the main algorithm in C
print STDERR "Calculating tree...\n";
system($scriptpath."tbc $ARGV[0].tmp.matrix");

# unlink temporary file
unlink("$ARGV[0].tmp.matrix");
