#!/usr/bin/perl

##########################################################################################
#	  This file is part of proteinortho.
#	  (C) 2009 Marcus Lechner
# 
#	  proteinortho is free software; you can redistribute it and/or modify
#	  it under the terms of the GNU General Public License as published
#	  by the Free Software Foundation; either version 2, or (at your
#	  option) any later version.
#
#	  proteinortho is distributed in the hope that it will be useful, but
#	  WITHOUT ANY WARRANTY; without even the implied warranty of
#	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	  General Public License for more details.
#
#	  You should have received a copy of the GNU General Public License
#	  along with proteinortho; see the file COPYING.  If not, write to the
#	  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#	  Boston, MA 02111-1307, USA.	
##########################################################################################

##########################################################################################
# About
##########################################################################################
# proteinortho
# input fasta files with proteins
# output matrix with ortholog proteins
# 
# @author Marcus Lechner
# @email marcus@bioinf.uni-leipzig.de
# @company Bioinformatics, University of Leipzig
# @version 3.1.6
# @date 19-10-2009
#
# based on proteinortho.pl version 2.2.0-beta from Sonja J. Prohaska
##########################################################################################

##########################################################################################
# Imports
##########################################################################################
use FindBin ();
use lib "$FindBin::Bin";
use strict;
use warnings "all";
use File::Basename;
use threads;
use threads::shared;
use vars qw/@query $cpu $debug $blastoutpath $force $verbose $fforce $keep $blastonly %seqs/;
use hashi_s2i;

# path to the C-part of this program
my @tmppath = fileparse($0);
our $scriptpath = $tmppath[1];
##########################################################################################

##########################################################################################
# Install signal handler
##########################################################################################
$SIG{'INT'}  = 'sighandler';
$SIG{'TERM'} = 'sighandler';
##########################################################################################

##########################################################################################
# Usage, defaults and parameters
##########################################################################################
my $usage= << "JUS";
  usage: proteinortho.pl [OPTIONS] MULTIFASTA1 MULTIFASTA2 [MULTIFASTA...]
  or:    proteinortho.pl [OPTIONS] LIST_OF_MULTIFASTAS
  needs: blastall, formatdb
version: 3.1.6
options: -e=        e-value [default: 1e-10]
         -a=        number of processors to use [default: 1]
         -p=        blast program {blastn|blastp} [default: blastp]
         -r=0|1     only *r*eciprocal blast hits [default: 1 (on)]
         -m=        minimum similarity of best blast hits (0..1) [default: 0.95 (nearly equal)]
	   -selfblast applys a blast of species A against A to detect paralogs more acurate
         -cmat      do not exclude *c*onnected *c*omponents with paralogs in output
         -plog      write *p*airwise *b*last hits to logfile
         -plog=     change logfile-name [default: pb.log]
         -clog      write *c*onnected *c*omponents to logfile
         -clog=     change logfile-name [default: cc.log]
         -ulog      write ultimate log: postprocess plog and clog to logfile
                    (very time intensive!)
         -ulog=     change logfile-name [default: ultimate.log]
         -f         force blastall (even if blast output is found)
         -ff        force formatdb (even if databases are found)
         -dir=      blast-output directory (default is the working directory)
         -remove    removes blastoutputs from blast-output directory after use
         -debug     keeps temporary files for debugging
         -blastonly only do the blasts: capable to run on multiple machines
                    syncronized via file in blast-output directory, clean break
                    with 'touch stop' or SIGTERM (kill pid of the script)
         -blastdone use after a complete run with blastonly on huge datasets
                    does not check fasta or blastfiles, just runs the algorithm
         -verbose   gives detailed information about what happens

         		 only one entry per in line in LIST_OF_MULTIFASTAS
                   DIFFERENT LOGFILES ARE RECOMMANDED!

-----------------------------------------------------------------------------
IMPORTANT! ensure that your protein IDs are unique ACROSS species! IMPORTANT!
IMPORTANT!   (eventually append species IDs to your protein IDs)   IMPORTANT!
-----------------------------------------------------------------------------
JUS

my @options = grep {substr($_, 0, 1) eq '-'} @ARGV;
my @others  = grep {substr($_, 0, 1) =~ /^[^-]/} @ARGV;

## default settings:
my $evalue = "1e-10";
my $blast = "blastp";
my $clog; #logfile
my $clogref;
my $plog; #logfile
my $plogref;
my $ulog; #logfile
my $ulogref;
my $cmat = 0;
my $noselfb = 1;
my $maxl = 0.95;
my $reciprocal = 1;
my $bhold = 1;
my $blastdone = 0;
our @joinable = (); # 4threading
share(@joinable);   # 4threading
$cpu = 1;           # 4threading
share($cpu);
$debug = 0;
$keep = 1;
$force=0;
$fforce=0;
$verbose=0;
$blastonly = 0;
our $gotlock = 0;	# Makes shure lock wont be stolen from another process

## setting parameters:
foreach my $option (@options){
  if    ($option =~ m/^-remove$/)     { $keep = 0;   }
  elsif ($option =~ m/^-blastonly$/)  { $blastonly = 1;   }
  elsif ($option =~ m/^-blastdone$/)  { $blastdone = 1; }
  elsif ($option =~ m/^-verbose$/)    { $verbose = 1;  }
  elsif ($option =~ m/^-debug$/)      { $debug = 1;  }  
  elsif ($option =~ m/^-selfblast$/)  { $noselfb = 0;  }  
  elsif ($option =~ m/^-f$/)          { $force = 1;  }
  elsif ($option =~ m/^-ff$/)         { $fforce = 1; }
  elsif ($option =~ m/^-r=([01])$/)   { $reciprocal = $1; }
  elsif ($option =~ m/^-p=(.*)$/)     { $blast = $1; }
  elsif ($option =~ m/^-e=(.*)$/)     { $evalue= $1; }
  elsif ($option =~ m/^-a=(\d*)$/)    { $cpu   = $1; }
  elsif ($option =~ m/^-m=(\d+.?\d*)$/)    { $maxl  = $1; }
  elsif ($option =~ m/^-cmat$/)       { $cmat  = 1; }
  elsif ($option =~ m/^-plog$/)       { $plog  = "pb.log"; }
  elsif ($option =~ m/^-plog=(.*)$/)  { $plog  = $1; }
  elsif ($option =~ m/^-clog$/)       { $clog  = "cc.log"; }
  elsif ($option =~ m/^-clog=(.*)$/)  { $clog  = $1; }
  elsif ($option =~ m/^-ulog$/)       { $ulog  = "ultimate.log"; }
  elsif ($option =~ m/^-ulog=(.*)$/)  { $ulog  = $1; }
  elsif ($option =~ m/^-dir=(.*)$/)   { $blastoutpath  = $1; }
  else  { die "Invalid command line option: \'$option\'!\n$usage"; }
}

# a list can also be given
# if only one argument is given
if (scalar(@others)==1) {
	# open the (possible) list 
	open(LIST,"<$others[0]") || die ("Could not find MULTIFASTA-list. \n$usage");
	@others = <LIST>;
	chomp(@others);
	close(LIST);
	# use it's lines as arguments
	foreach my $line (@others) {
		unless ($line =~ /.\../) {
			die ("$line in $others[0]\n does not seem to be a file!\nAt least two sequence files expected: MULTIFASTA1 MULTIFASTA2!\n\n$usage");
		}
	}
}

# still we need at least two
die "At least two sequence files expected: MULTIFASTA1 MULTIFASTA2!\n$usage"
    unless scalar(@others)>=2;

# bad parameter
if ($maxl < 0 || $maxl > 1) {
	die "Value for parameter m must be within 1 and 2\n$usage";
}

# bad parameter combination
if ($blastonly == 1 && $keep == 0) {
	die "Combination of remove and blastonly does not make much sense\n$usage";
}

# check file accessibility
print STDERR "Checking accessibility of files... \n";
for (my $i=0; $i<scalar(@others); $i++) {
  die "Can\'t read file \'$others[$i]\'!\n"
      unless open(FH, $others[$i]);
  close FH;
}

# initializing filehandles
if ($plog) {
  die "Can\'t open/write to file \'$plog\'!\n"
      unless open (PLOG, ">$plog");
  $plogref = \*PLOG;
}
if ($clog) {
  die "Can\'t open/write to file \'$clog\'!\n"
      unless open (CLOG, ">$clog");
  $clogref = \*CLOG;
}
if ($ulog) {
  unless ($plog ne '' & $clog ne '') {
    die "you have to set \'-plog -clog\' in combination with \'-ulog\'!\n";
  }
  die "Can\'t open/write to file \'$ulog\'!\n"
      unless open (ULOG, ">$ulog");
  $ulogref = \*ULOG;
}

##########################################################################################
# Main programm
#
##########################################################################################

##########################################################################################
# Step 1: format input-databases
##########################################################################################
my $stat_runs = scalar(@others)*(scalar(@others)+1)/2;		# 3.1.5

unless ($noselfb) {
	$stat_runs += scalar(@others);		# 3.1.6
}

my $stat_counter = 0;
my $glob_counter = 0;
my $stime;

# don't do if blast was already done
unless ($blastdone) {
print STDERR "Formating databases...\n";

# for each given fasta-file
for (my $i=0; $i<scalar(@others); $i++) {
  if ($force == 1 || do_formatdb($others[$i],$blast) == 1) {
	# Wait for free Thread-Slots
	my @tliste = threads->list();
	while(defined($tliste[0]) && scalar@tliste >= $cpu*1.5) {
		sleep(1);
		if (scalar(@joinable) > 0) {
			lock(@joinable);

			# Fetch finished threads
			while(scalar@joinable>0) {
				my $tid = shift(@joinable);
				my $thr = threads->object($tid);
				$thr->join();
				@tliste = threads->list();
			}
		}
	}

	# formatdb
	my $dbthread = threads->create('formatdb', ($others[$i],$blast));
  }
  else {
	if ($verbose) {print STDERR $others[$i]."\n-> formated database exists.\n";}
  }
}

# Gather unfinished threads
foreach my $dbthread (threads->list()) {
	$dbthread->join();
}
@joinable = ();

##########################################################################################
# Step 2: run the blasts (the biggest part)
##########################################################################################

print STDERR "Doing the blasts...\n";

# define blastoutpath
unless (defined($blastoutpath)) {
	my $prepath=qx('pwd');
	chomp($prepath);
	$blastoutpath=$prepath;
}
# Remove slash in the end
$blastoutpath =~ s/\/$//; 


##begin collect pairwise blast hits for all species pairs:
my @all;                    #square matrix (#species x #species),
                            #holds lists of reciprocal best blast hits
                            #for all species pairs


# Syncronisation for blastonly
my $s_i = 0; # only needed for sync in blastonly case
my $s_j = 0;
if ($blastonly == 1) {
	# init syncfile
	unless (-e $blastoutpath."/sync") {
		print STDERR "Syncfile not found. Adding new one: ".$blastoutpath."/sync\n";
		&fileLock;
		open(SYNC,">$blastoutpath"."/sync") || die "sync: $!";
		print SYNC "0 0";
		close(SYNC);
		&fileUnlock;
		print STDERR "Sync creation done\n";
	}
	else {
		print STDERR "Syncfile found. Sycronizing...\n";
	}
	# read current progress from syncfile
	($s_i, $s_j) = &sync();
	$glob_counter = ($s_i-1*$s_i)/2+$s_j-$s_i;
}

# start the blastruns
$stime = time;
for (my $i=$s_i; $i<scalar(@others)-$noselfb; $i++) {			# -1
  for (my $j=$i+$noselfb; $j<scalar(@others); $j++) {			#+1
	$glob_counter++;

	if ($verbose && $stat_counter%20 == 0) {
  		my $takentime = time - $stime;
		my $ttg = ($takentime/$glob_counter*$stat_runs-$takentime)/60;
		my $days = int($ttg/(60*24));
		my $hours = int(($ttg-60*24*$days)/60);
		my $minutes = int($ttg-60*24*$days-60*$hours);
		printf STDERR ("%.2f%% blasted overall -> ", ($glob_counter/$stat_runs*100));
		print STDERR "I think $days days $hours hours $minutes minutes to go\n";
	}

	# Syncronisation for blastonly
	if ($blastonly == 1) { 
		# skip what others should have done
	  	if ($i < $s_i || ($i == $s_i && $j <= $s_j)) {next;}
		
		# tell all I do a the job
	    	($s_i, $s_j) = &sync($i,$j); 		
		# stop order received, don't start new runs
	    	if ($s_i == -1) {$i = scalar(@others);last;} 
		
		
	    	unless ($i > $s_i || ($i == $s_i && $j > $s_j)) {next;}	 # skip what others have
	    	$stat_counter++;
		if ($verbose && $stat_counter%20 == 0) {
			printf STDERR ("%.2f%% by this process \n", ($stat_counter/$stat_runs*100));
		}
	}

    ## Precheck if files already exsist, saves waiting for threadslots in this case
    if (&blast_needed($others[$i], $others[$j])) {

    # Wait for free Thread-Slots
    my @tliste = threads->list();
	while(defined($tliste[0]) && scalar@tliste >= $cpu) {
		sleep(1);
		if (scalar(@joinable) > 0) {
			lock(@joinable);
			# Fetch finished threads
			while(scalar@joinable>0) {
				my $tid = shift(@joinable);
				my $thr = threads->object($tid);
				$thr->join();
				@tliste = threads->list();
			}
		}
	}
	my ($blastthread) = threads->create('blast', ($others[$i], $others[$j], $blast, $evalue, $bhold, $maxl, $plogref, $reciprocal));
    } # if (&blast_needed)
    elsif ($verbose) {	
	    print STDERR "$others[$i] vs. \n$others[$j] \n-> blastall output exists.\n\n";
	    print STDERR "$others[$j] vs. \n$others[$i] \n-> blastall output exists.\n\n";
    }
  }

}

# Gather unfinished threads
foreach my $thr (threads->list()) {
	$thr->join();
}
@joinable = ();

  if ($s_i == -1) { print STDERR "\nCalculation terminated due to existance of stop-file\n";}

if ($blastonly == 1) {
  print STDERR "Blastoutput generated\n";
  exit;
}
} # blastdone
else {
	print STDERR "Running in mode bastdone. Blastoutput should already exist!\n";
}

##########################################################################################
# Step 3: write headers and prepare for calculation
##########################################################################################

## Invert name to array-index mapping 
for (my $j=0; $j<scalar(@others); $j++) {
	$seqs{$others[$j]}=$j;
}


## print output header (in a thread while programm runs further)
print STDERR "Printing output header...\n";
my $h_thread = threads->create('print_output_header', (\@others, *STDOUT));

## read blastdata, find best hit(s) und transform data as input for c-programm
print STDERR "Transforming pairwise blast data...\n";

# Gather information
our $map = hashi_s2i::s2i_map->new();
our $map_counter = -1;

# open needed files
open(COUT,">cout");
open(MAPPING,">cout.mapping");
open(GVMAPPING,">cout.gvmapping");

# no. of jobs to do per thread
my $thread_size = 25;
$cpu++; # (headerthread -> +1)

$stat_counter = 0;
$stime = time;

# for all species pairs
for (my $i=0; $i<scalar(@others)-$noselfb; $i++) {			### 3.1.2 scalar -1	3.1.5
  # define start and endvalue for the next thread-job
  my $min = $i+$noselfb;								# 3.1.2 $i+1  3.1.5
  my $max = 0;
    for (my $j=$i+$noselfb; $j<scalar(@others); $j++) {			## 3.1.1 +1  3.1.5
     # endvalue for the next thread-job
	$max = $j;

	# init-thread
      if ($max - $min > $thread_size || $max == scalar(@others)-1) {
		my $out = 0;
		# Wait for free Thread-Slots
		my @tliste = threads->list();
		while(defined($tliste[0]) && scalar(@tliste) >= $cpu) { 
			sleep(1);
			if (scalar(@joinable) > 0) {
			lock(@joinable);
	
				# Fetch finished threads
				while(scalar(@joinable)>0) {
					my $tid = shift(@joinable);
					my $thr = threads->object($tid);
					my @ergs = $thr->join();
					my $t_i = -1;
					# transform the gathered information
					foreach (@ergs) {
						# ergs = ref to an array ($t_i, $t_j, $pairs1, $pairs2)
						# will be stored in a file

					      # hash: protein in species i (key) protein in species j (value) (href1)
				            # hash: protein in species j (key) protein in species i (value) (href2)
						
						pairs2edges($_->[2],$_->[0],$_->[1]);
						# bidirectional links -> 2nd run uneeded
						# pairs2edges($_->[3],$_->[1],$_->[0]); 
						$stat_counter++;
						if ($stat_counter%20 == 0) {$out = 1;}
					}
					
					@tliste = threads->list();
				}

   			     # Stats
			     if ($verbose && $out) {
					my $takentime = time - $stime;
					my $ttg = ($takentime/$stat_counter*$stat_runs-$takentime)/60;
					my $days = int($ttg/(60*24));
					my $hours = int(($ttg-60*24*$days)/60);
					my $minutes = int($ttg-60*24*$days-60*$hours);
					printf STDERR ("%.2f%% transformed -> ", ($stat_counter/$stat_runs*100));
					print STDERR "I think $days days $hours hours $minutes minutes to go\n";
			     }
			}
		} # CPU-Slot is now free

		# run the recent jobs as thread	
		# find reciprocal best blast hits with pair_po
	
		my ($coordinator) = threads->create('mfa_coordinator', (\@others, $i, $min, $max, $blast, $evalue, $bhold, $maxl, $plogref, $reciprocal));
		# reset min for next thread
		$min = $j+1;

	} # init-thread
  }
}
$cpu--; # nomalize again

# Join Headerthread (makes sure it has finished now)
$h_thread->join();

# Gather unfinished threads
foreach my $thr (threads->list()) {
	my @ergs = $thr->join();
	# Ergebnisse auswerten
	foreach (@ergs) {
		pairs2edges($_->[2],$_->[0],$_->[1]);
#		pairs2edges($_->[3],$_->[1],$_->[0]);
	}
}
@joinable = ();

# free memory
$map->clear();
$map->DESTROY();
undef($map);

##########################################################################################
# Step 4: gather the connected components (in C)
##########################################################################################
# run the cc-finder algorithm in C
print STDERR "Seaching for connected components... \n";
system($scriptpath.'ccf >cout.ccs');

# reverse map, protein-nr. -> protein-id
our @imap = ();
# load from mapping file
open(MAPPING,"<cout.mapping");
while (<MAPPING>) {
	my @line = split;
	chomp(@line);
	$imap[$line[1]] = $line[0];
}
close(MAPPING);

# write log (threaded)
if ($plogref) {
	close($plogref);
	print STDERR "Writing pairwise blast hits to $plog... \n";
	$h_thread = threads->create('cat_plog');
}

##########################################################################################
# Step 5: transform the connected components for output
##########################################################################################

# to write mono-orthologous proteins to output
print STDERR "Writing sets of mono-orthologous proteins to std-output... \n";
# we open the c-output
open(CCS,"<cout.ccs");
while (<CCS>) {
	chomp;	
	my @rows = split;
	# if there are no paralogs, just print
	if ($rows[scalar(@rows)-1] ne "paralog") {
			# print to std
			print join("\t",&invert(@rows)),"\n";			
	}
	# else remove the paralog-tag
	else {
		pop(@rows);
		# if paralogs are wanted, print line too
		if ($cmat) {
			print join("\t",&invert(@rows)),"\n";
		}
	}
	
	

	# print the log if wanted
	if ($clogref) {
		my $c = 0;
		foreach (@rows) {
			if ($_ ne "*") {$c++;}
		}
		@rows = &invert(@rows);
		print $clogref "# $c\n";
		for (my $i = 0; $i < scalar(@rows); $i++) {
			my @genes = split(",",$rows[$i]);
			foreach (@genes) {
				unless ($_ eq "*") {print $clogref "$_ $i\n";}
			}
		}
	}

}
close(CCS);

undef(@imap);

##########################################################################################
# Step 6: do the logs, cleanup and finish
##########################################################################################

if ($plogref) {
	# Join plog-thread
	$h_thread->join();
}

if ($cmat) {
	print STDERR "Writing sets of co-orthologous genes... \n";
}

if ($clogref) {
	print STDERR "Writing connected components to $clog... \n";
}

close $clogref if $clogref;

unless ($debug) {
 unlink("cout");
 unlink("cout.mapping");
 unlink("cout.gvmapping");
 unlink("cout.ccs");
}

## begin compute ultimate log (this takes very very long)
#------------------------------------------------------------------------------
if ($ulog) {
  print STDERR "Writing ultimate log to $ulog... \n";
  open CLOG, "<$clog" or die "Can\'t read/open \'$clog\'!\n";
  my $k;
  my $k_wc;
  my $out;
  while (<CLOG>) {
    if ($_ =~ m/^\#\s*(\d+)/) {
      $k_wc=$k=$1;
      next;
    }
    elsif ($_ =~ m/^\s*$/) {
      next;
    }
    else {
      $k-=1;
      my @line=split;
      $out .= `grep "$line[0]" $plog`;
      if ($k == 0) {
	open TMP, ">tmp" or die  "Can\'t open/write to \'tmp\'!\n";
	print TMP $out;
	close TMP;
	my $sort=`sort tmp | uniq`;
	my $l +=`sort tmp | uniq | wc -l`;
	print $ulogref "# $l $k_wc\n$sort";
	$out='';
      }
    }
  }
  close $ulogref;
}
#------------------------------------------------------------------------------
#end compute ultimate log

print STDERR "Finished.\n";


##########################################################################################
# FINISH, we are done
##########################################################################################





























##########################################################################################
# Subroutines
#
##########################################################################################

##########################################################################################
# Fasta-Format
##########################################################################################

# checks if formatdb is needed (unneeded, if the files already exist)
# make shure the path-detecion is the same as in sub formatdb
sub do_formatdb {
  my($faa,$blast)=@_;
#	print STDERR "File: $faa\n";
  my $val=1;
  if ($blast eq "blastn") {
    if (-e "$faa.nhr" &&
	-e "$faa.nin" &&
	-e "$faa.nsq") {
      $val=0;
    }
  }
  elsif ($blast eq "blastp") {
    if (-e "$faa.phr" &&
	-e "$faa.pin" &&
	-e "$faa.psq") {
      $val=0;
    }
  }
  else {
    die "$blast is not supported yet! Feel free to adapt the code to your needs. The programmer.\n"
  }
  return $val;
}

# Run format-db (THREAD)
sub formatdb {
  my($faa,$blast)=@_;
##(1) format multi-fasta files with formatdb
#------------------------------------------------------------------------------
  if ($blast =~ m/^blastp$/i) {
    if ($verbose) {print STDERR "Formating database $faa...\n";}
    `formatdb -i $faa -p T -o F`;
  #  print STDERR "done\n";
  }
  elsif ($blast =~ m/^blastn$/i) {
    if ($verbose) {print STDERR "Formating database $faa...\n";}
    `formatdb -i $faa -p F -o F`;
   # print STDERR "done.\n";
  }
  else {
    die "$blast is not supported yet! Feel free to adapt the code to your needs. The programmer.\n"
  }
#------------------------------------------------------------------------------

  # note, this thread has finished
  lock(@joinable);
  push(@joinable,threads->tid());
  return;
}

##########################################################################################
# Blast
##########################################################################################
# checks if a blast is needed (unneeded, if the files already exist)
# make shure the path-detecion is the same as in sub blast
sub blast_needed {
  my ($faa1, $faa2)=@_;
  my ($path1,$name1)=file2path_name($faa1);
  my ($path2,$name2)=file2path_name($faa2);

  my $blastout1=$blastoutpath."/".$name1.".".$name2.".bla";
  my $blastout2=$blastoutpath."/".$name2.".".$name1.".bla";

  if ($force == 0 && -e "$blastout1" && -e "$blastout2") {
	return 0;
  }
  return 1;
}

# prepares and runs the blasts (THREAD)
sub blast {
  my ($faa1, $faa2, $blast, $evalue, $bhold, $maxl, $plogref, $reciprocal)=@_;

  # gather paths
  my ($path1,$name1)=file2path_name($faa1);
  my ($path2,$name2)=file2path_name($faa2);
 
  #do blastall: faa1 as query, faa2 as database
  my $blastout1=$blastoutpath."/".$name1.".".$name2.".bla";
  if ($verbose) {print STDERR "$faa1 vs.\n$faa2:\ngenerating blastall output...\n";}
  blastall($faa1,$faa2,$blast,$evalue,$blastout1);

  unless ($faa1 eq $faa2) {		# 3.1.1 bidirectional not needed
	  #do blastall: faa2 as query, faa1 as database
	  my $blastout2=$blastoutpath."/".$name2.".".$name1.".bla";
	  if ($verbose) {print STDERR "$faa2 vs.\n$faa1:\ngenerating blastall output...\n";}
	  blastall($faa2,$faa1,$blast,$evalue,$blastout2);
  }

  # note, this thread has finished
  lock(@joinable);
  push(@joinable,threads->tid());

  return;
}


# does the blast and writes it to harddrive
sub blastall {
  my ($query, $db, $blast, $evalue, $out)=@_;
  `blastall -d $db -i $query -p $blast -a 1 -m8 -e $evalue >  $out`;
}

##########################################################################################
# Blast reading and analysis
##########################################################################################
# runs mfa2pairs in groups (j = min...max) (THREAD)
# therefore costs for threadscreation are lower
sub mfa_coordinator {
# \@others, $i, $min, $max, $blast, $evalue, $bhold, $maxl, $plogref, $reciprocal
	my $othersref = shift;
	my $i = shift;
	my $min = shift;
	my $max = shift;
	my @ergs = ();
	# run the jobs seperately
	for (my $j = $min; $j <= $max; $j++) {
	    my @erg = &mfa2pairs($i,$j,$othersref->[$i], $othersref->[$j], @_);
	    # store the results
	    push(@ergs,\@erg);
	}

	# note, this thread has finished
	lock(@joinable);
	push(@joinable,threads->tid());

	# return the results
	return @ergs;
}


# transformes blastoutput to hash-maps with the given criteria
sub mfa2pairs {
  my ($t_i, $t_j, $faa1, $faa2, $blast, $evalue, $bhold, $maxl, $plogref, $reciprocal)=@_;
  my ($path1,$name1)=file2path_name($faa1);
  my ($path2,$name2)=file2path_name($faa2);

  # Get blastout location
  my $blastout1=$blastoutpath."/".$name1.".".$name2.".bla";
  my $blastout2=$blastoutpath."/".$name2.".".$name1.".bla";

  # process blast output 1 & 2
  my $pairs1=blastout2pairs($blastout1, $evalue, $bhold, $maxl);
  my $pairs2=blastout2pairs($blastout2, $evalue, $bhold, $maxl);

  # remove Blast-Output
  if ($keep == 0) {
	unlink($blastout1);
	unlink($blastout2);
  }

  # delete directed edges (non-reciprocal hits) if this is what the user wants
  if ($reciprocal) {
	reciprocal_pairs($pairs1,$pairs2);
  }

  # logging (threaded, so evey entry in a special file)
  if ($plogref) {
    open(my $plogref2,">logpart_".$$."_".$t_i."_".$t_j.".tmp");
    print_pairs($pairs1,$faa1,$faa2,$plogref2);
    print_pairs($pairs2,$faa2,$faa1,$plogref2);
    close($plogref2);
  }

  #convert directed to undirected links (a programming requirement)
  unless ($reciprocal) {
    complete_pairs($pairs1,$pairs2);
  }

  # for all pairs are undirected, we only need one list
  return($t_i,$t_j,$pairs1);
}

#CAUTION! this subroutine modifies %pairs!
#called by mfa2pairs, checks if all pairs are reciprocal
#kicks them otherwise
sub reciprocal_pairs {
  my ($pairs1,$pairs2)=@_;

  foreach my $i1 (keys(%$pairs1)) {
    foreach my $j1 (keys(%{$pairs1->{$i1}})) {
      #if link is unidirectional
      unless (exists($pairs2->{$j1}) && exists($pairs2->{$j1}->{$i1})) {
	#delete link and nodes loosing all outgoing links
	_delete_pairs($pairs1,$i1,$j1);
      }
    }
  }
  foreach my $i2 (keys(%$pairs2)) {
    foreach my $j2 (keys(%{$pairs2->{$i2}})) {
      unless (exists($pairs1->{$j2}) && exists($pairs1->{$j2}->{$i2})) {
	#delete link and nodes loosing all outgoing links
	_delete_pairs($pairs2,$i2,$j2);
      }
    }
  }
}

#CAUTION! this subroutine modifies %pairs!
#called by reciprocal_pairs
sub _delete_pairs {
  my ($pairs,$ikey,$jkey)=@_;
  #delete link
  delete($pairs->{$ikey}->{$jkey});
  #if degree of the node is now zero
  if (scalar(keys(%{$pairs->{$ikey}}))<1) {
    #delete node
    delete($pairs->{$ikey});
  }
}

#CAUTION! this subroutine modifies %pairs!
#called by mfa2pairs, makes all pairs bidirectional
sub complete_pairs {
  my ($pairs1,$pairs2)=@_;
  foreach my $i1 (keys(%$pairs1)) {
    foreach my $j1 (keys(%{$pairs1->{$i1}})) {
      unless (exists($pairs2->{$j1}->{$i1})) {
	$pairs2->{$j1}->{$i1}=[$pairs1->{$i1}->{$j1}->[0],
			       $pairs1->{$i1}->{$j1}->[1],
			       $pairs1->{$i1}->{$j1}->[2]];
      }
    }
  }
  foreach my $i2 (keys(%$pairs2)) {
    foreach my $j2 (keys(%{$pairs2->{$i2}})) {
      unless (exists($pairs1->{$j2}->{$i2})) {
	$pairs1->{$j2}->{$i2}=[$pairs2->{$i2}->{$j2}->[0],
			       $pairs2->{$i2}->{$j2}->[1],
			       $pairs2->{$i2}->{$j2}->[2]];
      }
    }
  }
}

#  takes: blast output location (either file or reference to filehandle)
#         thresholds for e-value (ehold) and bit score (bhold)
#         maximum number of links (maxl)
#returns: reference to hash:
#         key = gene in query
#         value = hash: key(s) = gene(s) in target
#                       value(s) = [0]=e-val, [1]=bit score, [2]=%ID
sub blastout2pairs {
  my ($blastout, $ehold, $bhold, $maxl)=@_;
  my %pairs;

  #$blastout is either a reference to a filehandle or a filename
  #BEGIN: prepair to read from whatever it is:
  my $fhref;
  #there is a blast output in a filehandle
  if (ref($blastout) eq "GLOB") { 
    $fhref=$blastout;
  }
  #there is a blast output in a file
  else {                          
    open($fhref,"<$blastout") or die "Can\'t open/read \'$blastout\'!\n";
  }
  #END: prepair

  #what to write to %pairs (return value)
	
  while (<$fhref>) {
#    0         1           2              3               4        5       6       7         8      9        10        11    
# Query id, Subject id, % identity, alignment length, mismatches, gap, q. start, q. end, s. start, s. end, e-value, bit score
	my @line=split;			#[0]=seq1_query(-i) [1]=seq2_db(-d)
                    			#[2]=%ID [10]=e-value [11]=bitscore

	if ($line[2] < 25) {next;}		# skip id < 25%, 3.1.3

	# query end and db end need to be less than 2x alnlength 3.1.5
	if ($line[9] > 2*$line[3] || $line[7] > 2*$line[3]) {
#		print STDERR "$line[9] || $line[7] > 2x $line[3]\n";
		next;
	}

    #hash structure only allows one one occurence
    #per pair of gene_i(query) and gene_d(database)
    #therefore, how should multiple occurences be treated?
    if (exists(${$pairs{$line[0]}}{$line[1]})) {
      #chose best scoring entry:
      #best scoring in terms of bitscore
      if ($pairs{$line[0]}->{$line[1]}->[2] >= $line[11]) {
	#better scoring line for this pair is already in %pairs
      }
      else {
	#write better scoring line to %pairs
	_write2pairs(\%pairs,\@line,$ehold,$bhold,$maxl);
      }
    }
    else {
      #write line to %pairs
      _write2pairs(\%pairs,\@line,$ehold,$bhold,$maxl)
    }
  }

   unless (ref($blastout) eq "GLOB") {
	close($fhref);
   }

  return \%pairs;
}

#CAUTION! this subroutine modifies %pairs!
#called by blastout2pairs
sub _write2pairs {
  my ($pairs, $line, $ehold, $bhold, $maxl) = @_;
# scientific notation of $ehold and $line->[10]: '>=' comparison does not work

	#gather best value
	my $best = -1;
	foreach (keys %{$pairs->{$line->[0]}}) {
#		print "\n$_ -> ";
#		print $pairs->{$line->[0]}->{$_}->[1];
#		print "\n";

		if ($best < $pairs->{$line->[0]}->{$_}->[1]) {
			$best = $pairs->{$line->[0]}->{$_}->[1];
		}
	}

	my $new = $line->[11];
#	print STDERR "$best vs $new = ".(($best+$new)/$best-1)." ($maxl)"; 
	if ($line->[0] ne $line->[1] && $line->[11] >= $bhold && ($best == -1 || ($best+$new)/$best-1 >= $maxl)) {	# 3.1.4 avoid selfhits, 3.1.1 adaptive choice
#		print STDERR "\nOK\n";
	#overwrite existing entry
	$pairs->{$line->[0]}->{$line->[1]}=[$line->[10],$line->[11],$line->[2]];
    	return "1";
	}	
	else {
#		print STDERR "\nNOK\n";	
	    return "0";
	}
}


##########################################################################################
# Output
##########################################################################################
# takes: list of filenames (species) and filehandle
# returns: nothing
# writes to filehandle
sub print_output_header {	#FOUTPUT return;
  my ($filesref,$out)=@_;
  # (1) print species names (column heads for table)
  print $out "# " . join("\t", @$filesref) . "\n";

  # (2) get and print number of sequences in the species file
  print $out "# ";

  foreach my $i (@$filesref) {
    my $wc = `grep ">" $i | wc -l`;
    my @line=split(" ",$wc);
    print $out "$line[0]\t";

  }
  print $out "\n";
  # Cleanup for this thread might wait a while until it get's joined
  undef($out);
}

# transforms pairs from hash-tables to edgelists
# and makes the IDs to numbers, so C can handle it more efficent
sub pairs2edges {
	my $pairs = shift;
	my $viech1 = shift;
	my $viech2 = shift;
	# foreach startnoed
      foreach my $key (keys %{$pairs}) {			
		# skip empty pairs
		if (scalar(keys %{$pairs->{$key}}) == 0) {next;}	# 3.1.4

		# if there is no mapping-entry yet
		unless ($map->has_key($key)) {
			# use the next number and store it
			$map->set($key,++$map_counter);
			# also store the mapping in files for reconstruction
			print MAPPING "$key\t".$map->get_int($key)."\n";
			print GVMAPPING $map->get_int($key)." $viech1\n";
		}
		
		# print the edge start
		print COUT $map->get_int($key);

		# foreach target node
		foreach my $subkey (keys %{$pairs->{$key}}) {	
			# if there is no mapping-entry yet
			unless ($map->has_key($subkey)) {
				# use the next number and store it
				$map->set($subkey,++$map_counter);
				# also store the mapping in files for reconstruction
				print MAPPING "$subkey\t".$map->get_int($subkey)."\n";
				print GVMAPPING $map->get_int($subkey)." $viech2\n";
			}	
			# print the edge target (could be more than one)
			print COUT " ",$map->get_int($subkey);
		}
	# finish the entry
	print COUT "\n";
	}
}

##########################################################################################
# Logging
##########################################################################################

# prints all pairs found within a mapping
sub print_pairs {
  my ($pairs,$i,$j,$to)=@_;
  my $iID=$seqs{$i};
  my $jID=$seqs{$j};
  my $nr_pairs;
  my $hold;
  foreach my $ikey (keys(%$pairs)) {
    foreach my $jkey (keys(%{$pairs->{$ikey}})) {
      $nr_pairs++;
      $hold .= sprintf "$ikey $iID $jkey $jID " . ${$pairs->{$ikey}->{$jkey}}[0] . " " .
      ${$pairs->{$ikey}->{$jkey}}[1] . "\n";
    }
  }
  print $to "# $nr_pairs $iID $jID\n";
  print $to $hold;
}

# cats log-files from the differnet threads together
sub cat_plog {
	# this way the seperated logfiles from the threads are builded:
      # "logpart_"$pid."_".$t_i."_".$t_j.".tmp

	# cat them together
	system("cat logpart_".$$."*.tmp >$plog");
	# and erase the parts afterswards
	system("rm -f logpart_".$$."*.tmp");
}

##########################################################################################
# Tools
##########################################################################################

# inverts lists of protein-numbers to their previous names
# mapped within imap
# used to transform the c-core output
sub invert {
	my @list = @_;
	foreach my $data (@list) {		
		# more than one protein found: resolve them with an extra run
		if ($data =~ /,/) {
			$data = join(",",&invert(split(/,/,$data)));
		}
		else {
			# nothing (*) stays nothing (*)
			unless ($data eq "*") {
				$data = $imap[$data];		
			}
		}
	}
	
	# now the list contains the previous names, not the numbers
	return @list;
}

# divides a string into filename and it's path
sub file2path_name {
  my $file =shift;
  my ($name, $path) = fileparse($file);
  return $path, $name;
}

##########################################################################################
# Syncronising (used only if blastonly parameter is set)
##########################################################################################

# tells if blast-job could done next, or was already taken by another process
# basic syncronising method, does locking on it's own
sub sync {
    # the requested job
    my $i = shift;	
    my $j = shift;

    # if the (file based) stop-signal is found, we will stop
	if (-e $blastoutpath."/stop") {
		print STDERR "Recived order to terminate procedure...\n Finishing running threads\n";
		return(-1,-1);
	}

      # lock
	&fileLock;

	# open the syncfile
	open(SYNC,"+<","$blastoutpath"."/sync") || die "sync: $!";
	my @content = <SYNC>;
	chomp(@content);
	# gather current job
	(my $s_i, my $s_j) = split(/ /,$content[0],2);
	if (defined($i) && defined($j)) {
		# if we do the job, update sync-data	
		if ($i > $s_i || ($i == $s_i && $j > $s_j)) {
			seek(SYNC,0,0);
			truncate(SYNC, 0);
			print SYNC "$i $j";
			
			# inform the user
			if ($verbose) {print STDERR "Taking Job No $i x $j\n";}
		}
	}
	close(SYNC);

	# unlock	
	&fileUnlock;

	# return the next job to be done
	return ($s_i, $s_j);
}

# locks the sync-file
# always has to be done, before reading or writing it
# don't forget to unlock afterwards
sub fileLock {
	my $counter = 0;
	# if there already is a lock, wait
      # for flock does not work within NFS, we simulate it
      # with a directory to stay comaptible
	while (!mkdir($blastoutpath."/lock",0755)) {   
		# sleep for 1 sec and try again
		sleep(1);     

		# if access failed to often, there might be a problem
		$counter++;
		if ($counter > 30) {
			die("Termination due to locking problem. Directory $_[0]");
		}
     }

     # note globaly, that we've got the lock
     $gotlock = 1;

     # if the locking took longer, note it within the error-stream
     if ($verbose && $counter > 0) {
	   	print STDERR "Got lock on sync after $counter trys\n";
     }
}

# unlocks the sync-file
sub fileUnlock {
	# only remove the lock, if it's mine
	if ($gotlock == 0) {return;}	 

	# remove file lock dir
	rmdir($blastoutpath."/lock");     

	# note globaly, that we don't have the lock anymore
      $gotlock = 0;
}

# is called when the signals int term are caught
# shuts down the process safely, even if more of them are running syncronized
# finishes all threads, removes it's lock and exits
sub sighandler {
    # read signal number
    my $signal = shift;

    # reinstall signal-handler
    $SIG{'INT'}  = 'sighandler';
    $SIG{'TERM'} = 'sighandler';

    # prevent termination procedure to run twice     
    if ($cpu == 0) {
	print STDERR "Finishing dirty. Temporary data may not be reusable and should be recalculated.\n";
	print STDERR "You might need to remove it manualy or use the force-option for the next run!\n";
	return;
    }

    # tell the user, we are going to shut down
    print STDERR "Signal: $signal caught. Sending the signal again finishes dirty!\nPrepering for clean termination.\nFetching active threads...\n";

    # don't start new threads while waiting
    $cpu = 0;

    # remove the file lock if you have one
    if ($gotlock == 1) {
    	&fileUnlock;
    }

    # fetch all threads
    # blastoutputs and fasta-databases are written within the threads
    # for everything else the return values do not matter at all if the process is canceled
    if (defined(threads->list())) {
	    foreach my $thread (threads->list()) {
			my @list = threads->list();
	   	      print STDERR " ".(scalar(@list))." threads running, please wait\n";
			$thread->join();
	    }
    }

    # tell the user, we had shut down successfuly
    print STDERR "\nRun aborted.\n";
    exit;
}

